@@ -596,6 +596,8 @@ class Evaluator:
         original_image_direction = self.sample_metadata[input_name]["original_direction"]
         original_image_origin = self.sample_metadata[input_name]["original_origin"]
         original_image_spacing = self.sample_metadata[input_name]["original_spacing"]
+        if image_array.shape[-1] == 1 and len(original_image_direction) > len(image_array.shape):
+            image_array = np.squeeze(image_array, axies=-1)
         img = sitk.GetImageFromArray(image_array)
         img.SetDirection(original_image_direction)
         img.SetOrigin(original_image_origin)
@@ -624,13 +626,13 @@ class Evaluator:
 
     def get_sample_result_from_patch_results(self, patch_results):
         sample_results = {}
-        sample_names = None
+        sample_names = {}
         for i_dataset_name, i_dataset_generator in self.data_generator.items():
             i_patch_results = patch_results[i_dataset_name]
             sample_results[i_dataset_name] = {}
             file_locations = np.asarray(i_dataset_generator.sample_locations)
 
-            sample_names, sample_indices = self._find_sample_names_from_patch_names(
+            sample_names[i_dataset_name], sample_indices = self._find_sample_names_from_patch_names(
                 i_dataset_generator
             )
 
@@ -764,14 +766,14 @@ class Evaluator:
 
                 _, sample_labels = self.get_sample_labels_from_patch_labels()
                 results_df = self.make_dataframe(
-                    sample_names,
+                    sample_names[i_dataset_name],
                     sample_predictions[i_dataset_name],
                     sample_labels[i_dataset_name],
                 )
                 results_df.to_csv(out_file, index=False)
 
                 self.write_image_predictions_to_files(
-                    sample_names, sample_predictions[i_dataset_name], None,
+                    sample_names[i_dataset_name], sample_predictions[i_dataset_name], None,
                 )
 
     def make_metric_dataframe(self, metrics: dict) -> pd.DataFrame:
